function [ normval, fopt, info ] = hinorm( E, A, B, C, D, options )
%
% PURPOSE:
%
% To compute H-infinity-norm of an H-infinity transfer function
%
%   G(s) = C*(sE - A)^(-1)*B + D
%
% corresponding to a continuous-time linear time-invariant descriptor 
% system ( sE-A, B, C, D ). ALL matrices E, A, B, and C are assumed to be
% SPARSE. It is assumed that the number of inputs and ouputs is small
% compared to the system order.
% The H-infinity-norm is defined as
%
%   ||G||_Hinf := sup_{w real} ||G(iw)||_2.
%
% ARGUMENTS:
%
% Inputs:
%
% E, A, B, C, D: System matrices defining the transfer function.
% options      : Options struct (use default values if empty). The 
%                following options can be specified:
%    options.tolout  : tolerance on the relative change between two
%                      subsequent iterates in the outer iteration
%                      (default = 1e-6);
%    options.tolin   : tolerance on the relative change between two
%                      subsequent iterates in the inner iterations
%                      (default = 1e-3);
%    options.inftol  : tolerance for w for evaluating 
%                      lim_{w -> infinity} ||G(iw)||_2 (default = 1e-8);
%    options.maxitout: maximum number of outer iterations (default = 10);
%    options.maxitin : maximum number of inner iterations (default = 30);
%    options.npolinit: number of dominant poles computed in initial stage
%                      (default = 20);
%    options.npolin  : number of dominant poles computed in each inner
%                      iteration (default = 5);
%    options.ntest   : number of maximum tests in order to find the correct
%                      starting pole (giving the global optimizer)
%                      (default = 5);
%    options.shift   : initial shift for dominant pole algorithm
%                      (default = 1i);
%    options.realchk : number of most dominant poles used to check whether
%                      the norm is attained at zero (default = 15);
%    options.dommeas : specifies the choice of the dominance measure as
%                      follows:
%                      = 'LR': take the largest |residue|/|real(pole)| 
%                              magnitude;
%                      = beta: take the largest 
%                              |residue|*exp(beta*real(pole)) magnitude;
%                      (default = 100);
%    options.psabschk: specifies whether the structured pseudospectral
%                      abscissa is computed again for the final epsilon as
%                      follows:
%                      = 0: structured pseudospectral abscissa is not
%                           computed;
%                      = 1: structured pseudospectral abscissa is computed;
%                      (default = 0);
%    options.prtlevel: specifies the print level as follows:
%                      = 0: return no information;
%                      = 1: return minimal information;
%                      = 2: return full informations;
%                      (default = 0).
%
% Outputs:
%
% normval      : Computed H-infinity-norm.
% fopt         : Frequency at which the norm value is attained.
% info         : Info struct that contains the following information:
%    info.error      : contains an error indicator as follows:
%                      = 0: return without error;
%                      = 1: maximum number of iterations exceeded;
%                      = 2: computed H-infinity-norm smaller than 
%                           ||G(iw_0)||_2 for a test frequency w_0; 
%    info.pseudoabs  : contains the structured pseudospectral abscissa, if 
%                      options.psabschk == 1;
%    info.iterout    : number of outer iterations (without computing
%                      info.pseudoabs); 
%    info.iterin     : number of inner iterations (without computing
%                      info.pseudoabs);
%    info.time       : time needed to compute the result (without computing
%                      info.pseudoabs).
%
% REFERENCES:
%
% [1] P. Benner and M. Voigt:
%     A structured pseudospectral method for H-infinity-norm computation of 
%     large-scale descriptor systems, 2012.
%
% AUTHOR:
%
% Matthias Voigt, Max Planck Institute for Dynamics of Complex Technical
% Systems, Magdeburg, Germany, 06/25/2013.
%
% REVISIONS:
%
% -
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  HINORM 1.0 Copyright (C) 2013 Matthias Voigt 
%
%  This program is free software: you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation, either version 3 of the License, or
%  (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%% CHECK ARGUMENTS.
%
% Start timing.
%
tic;
%
% Check the number of arguments.
%
if ( nargin < 5 )
    error( 'HINORM requires at least 5 input arguments.' );
elseif ( nargin == 5 )
    options = [];
elseif ( nargin > 6 )
    error( 'HINORM requires at most 6 input arguments.' );
end
%
if ( nargout > 3 )
    error( 'HINORM requires at most 3 output arguments.' );
end
%
% Check dimensions.
%
[ n1, m1 ] = size( B );
[ p1, n2 ] = size( C );
[ p2, m2 ] = size( D );
%
if ( n1 ~= n2 || p1 ~= p2 || m1 ~= m2 )
    error( 'The dimensions of B, C, and D are not compatible.' );
else
    p = p1;
    m = m1;
end
%
[ ne1, ne2 ] = size( E );
[ na1, na2 ] = size( A );
if ( ne1 ~= ne2 )
    error( 'The matrix E must be square.' );
elseif ( na1 ~= na2 )
    error( 'The matrix A must be square.' );
elseif ( ne1 ~= na1 )
    error( 'The dimensions of E and A are not compatible.' );
elseif ( ne1 ~= n1 )
    error( 'The dimensions of E and A are not compatible with B, C, and D.' );
else
    n = n1;
end
%
% Check for sparsity.
%
if ( ~issparse( E ) )
    error( 'The matrix E must be sparse.' );
elseif ( ~issparse( A ) )
    error( 'The matrix A must be sparse.' );
elseif ( ~issparse( B ) )
    error( 'The matrix B must be sparse.' );
elseif ( ~issparse( C ) )
    error( 'The matrix C must be sparse.' );
end
%
%% SET OPTIONS.
%
% tolout: tolerance on the relative change between two subsequent iterates 
% in the outer iteration.
%
if isfield( options, 'tolout' )
    tolout = options.tolout; 
else % default
    tolout = 1e-6;
end
%
% inftol: tolerance for w for evaluating lim_{w -> infinity} ||G(iw)||_2.
%
if isfield( options, 'inftol' )
    inftol = options.inftol; 
else % default
    inftol = 1e-8;
end
%
% maxitout: maximum number of outer iterations.
%
if isfield( options, 'maxitout' )
    maxitout = options.maxitout; 
else % default
    maxitout = 10;
end
%
% npolinit: number of dominant poles computed in initial stage.
%
if isfield( options, 'npolinit' )
    npolinit = options.npolinit; 
else % default
    npolinit = 20;
end
%
% ntest: number of maximum tests in order to find the correct starting pole
% (giving the global optimizer).
%
if isfield( options, 'ntest' )
    ntest = options.ntest; 
else % default
    ntest = 5;
end
%
% shift: initial shift for dominant pole algorithm.
%
if isfield( options, 'shift' )
    shift = options.shift; 
else % default
    shift = 1i;
end
%
% realchk: number of most dominant poles used to check whether the norm is 
% attained at zero.
%
if isfield( options, 'realchk' )
    realchk = options.realchk; 
else % default
    realchk = 15;
end
%
% psabschk: specifies whether the structured pseudospectral abscissa is 
% computed again for the final epsilon.
%
if isfield( options, 'psabschk' )
    psabschk = options.psabschk;
    if ( psabschk ~= 0 && psabschk ~= 1 )
         error( 'The argument options.psabschk must be 0 or 1.' );
    end
else % default
    psabschk = 0;
end
%
% prtlevel: specifies the print level.
%
if isfield( options, 'prtlevel' )
    prtlevel = options.prtlevel;
    if ( prtlevel ~= 0 && prtlevel ~= 1 && prtlevel ~= 2 )
         error( 'The argument options.psabschk must be 0, 1, or 2.' );
    end
else % default
    prtlevel = 0;
end
%
%% COMPUTATIONS.
%
% Initilizations.
%
warning off all;
iterin = 0;
iterout = 0;
info.error = 0;
%
% If D ~= 0, extend the realization to make D = 0.
%
if ( norm( D, 2 ) ~= 0 )
    E = sparse( blkdiag( E, zeros( p ) ) );
    A = sparse( blkdiag( A, -speye( p ) ) );
    B = sparse( [ B; D ] );
    C = sparse( [ C, speye( p ) ] );
    %
    % Determine new n, m, and p.
    %
    n = size( A, 1 );
    m = size( B, 2 );
    p = size( C, 1 );
end
%
% Compute test frequencies. Set options for SAMDP.
%
optsamdp.yEx_scaling = 0;
optsamdp.strategy = 'LR';
optsamdp.nwanted = npolinit;
optsamdp.displ = 0;
%
[ poles, ~, rightev, leftev, ~ ] = samdp( A, E, B, C', zeros( m, p ), ...
    shift, optsamdp );
%
% Extract the poles with non-negative imaginary part.
%
posimag = find( imag( poles ) >= 0 );
poles = poles( posimag );
rightev = rightev( :, posimag );
leftev = leftev( :, posimag );
ndpol = numel( rightev( 1, : ) );
%
% Scale the eigenvectors such that y*Ex = 1.
%
for j = 1 : ndpol
    tmp = leftev( :, j )'*E*rightev( :, j );
    tmp = tmp/abs( tmp );
    rightev( :, j ) = rightev( :, j )*sqrt( tmp );
    leftev( :, j ) = leftev( :, j )*sqrt( tmp );
end
%
% Sort the computed dominant poles and the associated eigenvectors with
% respect to the largest |residue|/|real(pole)| magnitude.
%
contobs = zeros( 1, ndpol );
for j = 1 : ndpol 
    contobs( j ) = norm( C*rightev( :, j )*leftev( :, j )'*B, 2 )/ ...
        abs( real( poles( j ) ) );
end
[ dom, pos ] = sort( contobs, 'descend' );
poles = poles( pos );
rightev = rightev( :, pos );
leftev = leftev( :, pos );
%
% Evaluate lim_{w -> infinity} ||G(iw)||_2.
%
omega = abs( max( imag( poles ) ) );
if ( omega == 0 )
    omega = 1;
end
ressum = Inf;
while ( ressum > inftol )
    omega = 10*omega;
    ressum = 0;
    for j = 1 : ndpol
        ressum = ressum + dom( j )/( omega - imag( poles( j ) ) );
    end
end
svinf = max( svds( C/( 1i*omega*E - A )*B ) );
%
% Check whether the most dominant poles are all real. Then the 
% H-infinity-norm is attained at zero or infinity.
%
if ( norm( imag( poles( 1 : min( ndpol, realchk ) ) ) ) ...
        < 1e3*eps )
    %
    % The most dominant poles are all real. The H-infinity-norm is attained
    % at zero or infinity.
    %
    svzero = max( svds( -C/A*B ) );
    if ( svzero > svinf )
        normval = svzero;
        fopt = 0;
        info.time = toc;
        if ( prtlevel > 0 )
            %
            % Print info.
            %
            fprintf( 'The H-infinity-norm is attained for fopt = 0. The norm value is %g.\n', normval );
            fprintf( 'The runtime is %g seconds.\n', info.time );       
        end
    else
        normval = svinf;
        fopt = Inf;
        info.time = toc;
        if ( prtlevel > 0 )
            %
            % Print info.
            %
            fprintf( 'The H-infinity-norm is attained for fopt = Inf. The norm value is %g.\n', normval );
            fprintf( 'The runtime is %g seconds.\n', info.time );     
        end
    end
    %
    if ( psabschk == 1 )
        info.pseudoabs = NaN;
    end
    info.iterin = iterin;
    info.iterout = iterout;
    return
end
%
% Evaluate ||G(iw_0)||_2 at test frequencies w_0.
%
svmax = 0;
for j = 1 : ndpol
    sv = max( svds( C/( 1i*imag( poles( j ) )*E - A )*B ) );
    if ( sv > svmax )
        svmax = sv;
    end
end
%
% If svmax < svinf, the norm is still assumed to be attained at infinity.
%
if ( svmax < svinf )
    normval = svinf;
    fopt = Inf;
    if ( prtlevel > 0 )
        %
        % Print info.
        %
        fprintf( 'The H-infinity-norm is attained for fopt = Inf. The norm value is %g.\n', normval );
    end
    %
    if ( psabschk == 1 )
        info.pseudoabs = NaN;
    end
    info.iterin = iterin;
    info.iterout = iterout;
    return
end
%
% Find a pole that generates a positive structured pseudospectral abscissa.
% If such a pole is not found, take the one with the maximum structured
% pseudospectral abscissa.
%
ntest = min( ndpol, ntest );
epsilon = 1/svmax;
maxpsa = -Inf;
%
if ( prtlevel > 0 )
    %
    % Print info.
    %
    fprintf( 'Phase 1: Searching initial pole:\n' );
end
%
for j = 1 : ntest
    [ psa, poletmp, rightevtmp, leftevtmp, numit ] = pseudoabs( E, A, ...
        B, C, poles( j ), rightev( :, j ), leftev( :, j ), epsilon, ...
        options );
    iterin = iterin + numit;
    iterout = iterout + 1;
    %
    if ( prtlevel > 0 )
        %
        % Print info.
        %
        fprintf( '    %d. outer iteration: epsilon = %g, psa = %g.\n', j, ...
            epsilon, psa );
    end
    %
    if ( psa > maxpsa )
        maxpsa = psa;
        spole = poletmp;
        srightev = rightevtmp;
        sleftev = leftevtmp;
        if( maxpsa > 0 )
            break
        end
    end
end
psa = maxpsa;
psaold = Inf;
%
% Iteration over epsilon (OUTER ITERATION).
%
if ( prtlevel > 0 )
    %
    % Print info.
    %
    fprintf( 'Phase 2: Iteration over epsilon:\n' );
end
%
for j = 1 : maxitout
    %
    % Save the old epsilon and update by a Newton step.
    %
    epsilonold = epsilon;
    epsilon = epsilon - psa/( norm( C*srightev, 2 )*norm( B'*sleftev, ...
        2 ) );
    if ( abs( 1 - epsilonold/epsilon ) < tolout || ... 
             ( abs( psaold/psa ) < 1e1 && abs( psaold/psa ) > 1e-1 ) )
        %
        % The relative change between two subsequent iterates in the outer
        % iteration is too small or two strucutred pseudospectral abscissa
        % do not differ by at least one order of magnitude.
        %
        % Return results.
        %
        fopt = imag( spole );
        normval = max( 1/epsilon, max( svds( C/( 1i*fopt*E - A )*B ) ) );
        %
        % Check for errors.
        %
        if ( ( 1 - tolout )*svmax > normval )
            info.error = 2;
            if ( prtlevel > 0 )
            %
            % Print info.
            %
            fprintf( 'The computed norm value is smaller than the largest singular value at a test frequency. This value is %g.\n', svmax );
            end
        end
        %
        info.iterin = iterin;
        info.iterout = iterout;
        %
        % Stop timing.
        %
        info.time = toc;
        %
        if ( prtlevel > 0 )
            %
            % Print info.
            %
            fprintf( '\n' );
            fprintf( 'The H-infinity-norm is attained for fopt = %g. The norm value is %g.\n', fopt, normval );
            fprintf( 'The algorithm needed %g outer and %g inner iterations.\n', info.iterout, info.iterin );
            fprintf( 'The runtime is %g seconds.\n', info.time );
            fprintf( '\n' );        
        end
        %
        if ( psabschk == 1 )
            if ( prtlevel > 0 )
                %
                % Print info.
                %
                fprintf( 'Phase 3: Compute the final structured pseudospectral abscissa (optional):\n' );
            end
            [ psa, ~, ~, ~, ~ ] = pseudoabs( E, A, B, C, spole, ...
                srightev, sleftev, epsilon, options );
            info.pseudoabs = psa;
            if ( prtlevel > 0 )
                %
                % Print info.
                %
                fprintf( 'The final structured pseudospectral abscissa is %g.\n', psa );
            end
        end
        %
        return
    end
    %
    % Save psa for later usage.
    %
    psaold = psa;
    %
    [ psa, spole, srightev, sleftev, numit ] = pseudoabs( E, A, B, C, ...
        spole, srightev, sleftev, epsilon, options );
    %
    if ( prtlevel > 0 )
        %
        % Print info.
        %
        fprintf( '    %d. outer iteration: epsilon = %g, psa = %g.\n', j, ...
            epsilon, psa );
    end
    %
    % Check whether maxitout is exceeded.
    %
    if ( j == maxitout )
        info.error = 1;
        normval = NaN;
        fopt = NaN;
        info.time = toc;
        if ( prtlevel > 0 )
            %
            % Print info.
            %
            fprintf( 'The maximum number of outer iterations has been exceeded.\n' );
            fprintf( 'No results are returned.\n' );
        end
    end
    %
    iterin = iterin + numit;
    iterout = iterout + 1;
end
