function [ psa, pole, rightev, leftev, numit ] = pseudoabs( E, A, B, C, ...
    pole, rightev, leftev, epsilon, options )
%
% PURPOSE:
%
% To compute the structured epsilon-pseudospectral abscissa of the transfer
% function 
%
%   G(s) = C*(sE - A)^(-1)*B.
%
% The structured epsilon-pseudospectral abscissa is defined as the real
% part of the rightmost pole of all perturbed transfer function of the form
%
%   G_Delta(s) = C*(sE - (A + B*Delta*C))^(-1)*B,
%
% where ||Delta||_2 < epsilon.
%
% ARGUMENTS:
%
% Inputs:
%
% E, A, B, C   : System matrices defining the transfer function.
% pole         : Initial pole used to compute the rightmost pseudopole.
% rightev,
% leftev       : Right and left eigenvectors associated to the input pole.
% epsilon      : Perturbation level.
% options      : Options struct (use default values if empty). These are 
%                the same options as for HINORM. The following options can
%                be specified:
%    options.tolout  : tolerance on the relative change between two
%                      subsequent iterates in the outer iteration
%                      (default = 1e-6);
%    options.tolin   : tolerance on the relative change between two
%                      subsequent iterates in the inner iterations
%                      (default = 1e-3);
%    options.inftol  : tolerance for w for evaluating 
%                      lim_{w -> infinity} ||G(iw)||_2 (default = 1e-8);
%    options.maxitout: maximum number of outer iterations (default = 10);
%    options.maxitin : maximum number of inner iterations (default = 30);
%    options.npolinit: number of dominant poles computed in initial stage
%                      (default = 20);
%    options.npolin  : number of dominant poles computed in each inner
%                      iteration (default = 5);
%    options.ntest   : number of maximum tests in order to find the correct
%                      starting pole (giving the global optimizer)
%                      (default = 5);
%    options.shift   : initial shift for dominant pole algorithm
%                      (default = 1i);
%    options.realchk : number of most dominant poles used to check whether
%                      the norm is attained at zero (default = 15);
%    options.dommeas : specifies the choice of the dominance measure as
%                      follows:
%                      = 'LR': take the largest |residue|/|real(pole)| 
%                              magnitude;
%                      = beta: take the largest 
%                              |residue|*exp(beta*real(pole)) magnitude;
%                      (default = 'LR');
%    options.psabschk: specifies whether the structured pseudospectral
%                      abscissa is computed again for the final epsilon as
%                      follows:
%                      = 0: structured pseudospectral abscissa is not
%                           computed;
%                      = 1: structured pseudospectral abscissa is computed;
%                      (default = 0);
%    options.prtlevel: specifies the print level as follows:
%                      = 0: return no information;
%                      = 1: return minimal information;
%                      = 2: return full informations;
%                      (default = 0).
%
% Outputs:
%
% psa          : Structured pseudospectral abscissa.
% pole         : Pseudopole realizing the structured pseudospectral
%                abscissa.
% rightev,
% leftev       : Right and left eigenvectors associated to the output pole.
% numit        : Number of iterations.
%
% REFERENCES:
%
% [1] P. Benner and M. Voigt:
%     A structured pseudospectral method for H-infinity-norm computation of 
%     large-scale descriptor systems, 2012.
%
% AUTHOR:
%
% Matthias Voigt, Max Planck Institute for Dynamics of Complex Technical
% Systems, Magdeburg, Germany, 06/25/2013.
%
% REVISIONS:
%
% -
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  HINORM 1.0 Copyright (C) 2013 Matthias Voigt 
%
%  This program is free software: you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation, either version 3 of the License, or
%  (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%% SET OPTIONS.
%
% tolin: tolerance on the relative change between two subsequent iterates 
% in the inner iteration.
%
if isfield( options, 'tolin' )
    tolin = options.tolin; 
else % default
    tolin = 1e-3;
end
%
% maxitin: maximum number of inner iterations.
%
if isfield( options, 'maxitin' )
    maxitin = options.maxitin; 
else % default
    maxitin = 30;
end
%
% npolin: number of dominant poles computed in each inner iteration.
%
if isfield( options, 'npolin' )
    npolin = options.npolin; 
else % default
    npolin = 5;
end
%
% dommeas: specifies the choice of the dominance measure.
%
if isfield( options, 'dommeas' )
    if ( options.dommeas == 'LR' )
        dommeas = 'LR';
    else
        dommeas = 'exp';
        beta = options.dommeas;
    end
else % default
    dommeas = 'LR';
end
%
% prtlevel: specifies the print level.
%
if isfield( options, 'prtlevel' )
    prtlevel = options.prtlevel;
    if ( prtlevel ~= 0 && prtlevel ~= 1 && prtlevel ~= 2 )
         error( 'The argument options.psabschk must be 0, 1, or 2.' );
    end
else % default
    prtlevel = 0;
end
%
%% COMPUTATIONS.
%
% Initilizations.
%
numit = 0;
n = size( A, 1 );
m = size( B, 2 );
p = size( C, 1 );
%
% Set options for SAMDP.
%
optsamdp.yEx_scaling = 0;
optsamdp.nwanted = npolin;
optsamdp.strategy = 'LR';
optsamdp.displ = 0;
%
% Compute initial perturbation.
%
As = A;
tmp = leftev'*E*rightev;
tmp = tmp/abs( tmp );
rightev = rightev*sqrt( tmp );
leftev = leftev*sqrt( tmp );
tmpb = B'*leftev;
tmpc = rightev'*C';
L = sparse( ( sparse( B*tmpb )*sparse( tmpc*C ) )/ ...
    ( norm( tmpb, 2 )*norm( tmpc, 2 ) ) );
A = sparse( As + epsilon*L );
%
% Iteration.
%
for k = 1 : maxitin
    %
    % Compute some dominant poles.
    %
    [ poles, ~, rightevs, leftevs, ~ ] = samdp( A, E, B, C', ...
        zeros( m, p ), pole, optsamdp );
    numit = numit + 1;
    ndpol = numel( rightevs( 1, : ) );
    %
    % Sort the computed dominant poles and the associated eigenvectors with
    % respect to the specified dominance measure.
    %
    contobs = zeros( ndpol, 1 );
    if ( strcmp( dommeas, 'LR' ) )
        for j = 1 : ndpol
            contobs( j ) = norm( C*rightevs( :, j )*leftevs( :, j )'*B, ...
                2 )/abs( real( poles( j ) ) );
        end
    else
        for j = 1 : ndpol
            contobs( j ) = norm( C*rightevs( :, j )*leftevs( :, j )'*B, ...
                2 )*exp( beta*real( poles( j ) ) );
        end
    end
    [ ~, pos ] = sort( contobs, 'descend' );
    psa = real( poles( pos( 1 ) ) );
    if ( prtlevel > 1 )
        %
        % Print info.
        %
        fprintf( '        %d. inner iteration: psa = %g.\n', k, psa );
    end
    %
    % Check for convergence.
    %
    if ( k >= 3 && psaold > psa )
        %
        % Iteration cycles. Return the results from the previous iteration.
        %
        psa = psaold;
        rightev = rightevold;
        leftev = leftevold;
        %
        if ( prtlevel > 1 )
            %
            % Print info.
            %
            fprintf( '        Iteration cycles, take the old value.\n' );
        end
        %
        return
    elseif ( k >= 2 && ( abs( psaold - psa ) < tolin*abs( psa ) ) )
        %
        % Iteration stagnates. Return the results from the current 
        % iteration.
        %
        pole = poles( pos( 1 ) );
        rightev = rightevs( :, pos( 1 ) );
        leftev = leftevs( :, pos( 1 ) );
        tmp = leftev'*E*rightev;
        tmp = tmp/abs( tmp );
        rightev = rightev*sqrt( tmp );
        leftev = leftev*sqrt( tmp );
        if ( prtlevel > 1 )
            %
            % Print info.
            %
            fprintf( '        Iteration stagnates, pseudospectral abscissa found.\n' );
        end
        %
        return
    end
    %
    % Save the structured pseudospectral abscissa from the current     
    % iteration for later usage.
    %
    psaold = psa;
    %
    % Update the perturbation matrix.
    %
    pole = poles( pos( 1 ) ); 
    rightev = rightevs( :, pos( 1 ) );
    leftev = leftevs( :, pos( 1 ) );
    tmp = leftev'*E*rightev;
    tmp = tmp/abs( tmp );
    rightev = rightev*sqrt( tmp );
    leftev = leftev*sqrt( tmp );
    tmpb = B'*leftev;
    tmpc = rightev'*C';
    L = sparse( ( sparse( B*tmpb )*sparse( tmpc*C ) )/ ...
        ( norm( tmpb, 2 )*norm( tmpc, 2 ) ) );
    A = sparse( As + epsilon*L );
    %
    % Save the right and left eigenvectors from the current iteration for 
    % later usage.
    %
    rightevold = rightev;
    leftevold = leftev;
end